//=============================================================================
//
//  CLASS Accelerator
//
//=============================================================================

#ifndef RAYTRACER_ACCELERATOR_H
#define RAYTRACER_ACCELERATOR_H

//== INCLUDES =================================================================
#include "Types.h"
#include "SurfaceList.h"

//== CLASS DEFINITION =========================================================

/*!
	Basis-Klasse fuer Beschleunigungsdatenstrukturen.
	Die Beschleunigungsdatenstrukturen werden im Konstruktor aufgebaut
	Darueber hinaus muss jede Datenstruktur eine intersect-Methode
	zur Schnittberechnung mit der Szene enthalten
*/

class SurfaceList;

class Accelerator
{
//-------------------------------------------------------------- private data
protected:
	const SurfaceList *m_scene;
	
//-------------------------------------------------------------- constructors
public:
	//! standard constructor
	Accelerator(){};

	//! destructor
	virtual ~Accelerator()
	{
	};

//--------------------------------------------------------------- intersections
    virtual bool intersect( const Ray&    clRay, 
	   		        	    const double  t0, 
				            const double  t1, 
					        TracingContext&  tContext ) const = 0;

};
//=============================================================================
#endif // RAYTRACER_ACCELERATOR_H
//=============================================================================