//=============================================================================
//
//  CLASS TriangleSoup - IMPLEMENTATION
//
//=============================================================================

#define RAYTRACER_TRIANGLESOUP_CPP


//== INCLUDES =================================================================
#include "TriangleSoup.h"
#include "PhongShader.h"
#include <fstream>

   
//== IMPLEMENTATION ========================================================== 
TriangleSoup::TriangleSoup( ) : 
                     m_bAABBComputed(false),
	                 m_selected(false),
	                 m_drawnThisCycle(false)
{
	m_clVertexList.clear();
	m_clNormalList.clear();
	m_clTriangleList.clear();
	m_clAABBox.Reset();
}


   
//== IMPLEMENTATION ========================================================== 
void TriangleSoup::loadMaterials( const char* sFilename, std::vector<Shader *>   *materialList, std::list<MaterialNode> *materialNodes )
{

    char	ac_line[256];

	std::ifstream istr_in( sFilename );
	if ( !istr_in.is_open() )
	{
		std::cerr << "could not read " << sFilename << std::endl;
	}
	
	PhongShader *actShader = NULL;

	std::cerr << " reading Materials " << sFilename << std::endl;	
	while( istr_in && ( !istr_in.eof() ) && istr_in.getline( ac_line, 256 ) )
	{
        float   ar, ag, ab, sr, sg, sb, rr, rg, rb, dr, dg, db;
        float   refractiveIndex, shininess;
		// comment
		if ( ac_line[0] == '#' ) continue;
		
		//Load Material File
		if ( strncmp(ac_line, "newmtl ", 7 ) == 0 )
		{
			char matname[1024];
			if ( sscanf( ac_line, "newmtl %s", matname ) )
			{
			    //Put Last Shader to the MaterialList
			    if( actShader != NULL )
                {
                    if ( actShader->getRefractiveIndex() != 1.0
                        && actShader->getRefractiveColor()[0] == 0.0
                        && actShader->getRefractiveColor()[1] == 0.0
                        && actShader->getRefractiveColor()[2] == 0.0 )
                    {
                        actShader->setRefractiveColor( ColorType( 1.0 ) );
                    }
			        materialList->push_back( actShader );
                }
			    
				//Create new Shader
				actShader = new PhongShader();
				    
				MaterialNode node;
				node.m_shader = actShader; 
				strcpy( node.m_name, matname );
				materialNodes->push_back( node );
			}
		}
		else if ( strncmp(ac_line, "Ns ", 3 ) == 0 )
		{ 
		    sscanf( ac_line, "Ns %f", &shininess );
		    if ( actShader != NULL )
		        actShader->setShininess( shininess );
		}
		else if ( strncmp(ac_line, "Ka ", 3 ) == 0 )
		{ 
		    sscanf( ac_line, "Ka %f %f %f", &ar, &ag, &ab );
		    if ( actShader != NULL )
		        actShader->setAmbientColor( ColorType( ar, ag, ab ) );
		}
		else if ( strncmp(ac_line, "Kd ", 3 ) == 0 )
		{ 
		    sscanf( ac_line, "Kd %f %f %f", &dr, &dg, &db );
		    if ( actShader != NULL )
		        actShader->setDiffuseColor( ColorType( dr, dg, db ) );
		}
		else if ( strncmp(ac_line, "Ks ", 3 ) == 0 )
		{ 
		    sscanf( ac_line, "Ks %f %f %f", &sr, &sg, &sb );
		    if ( actShader != NULL )
		        actShader->setSpecularColor( ColorType( sr, sg, sb ) );
		}
		else if ( strncmp(ac_line, "Ni ", 3 ) == 0 )
		{ 
		    sscanf( ac_line, "Ni %f", &refractiveIndex );
		    if ( actShader != NULL )
		        actShader->setRefractiveIndex( refractiveIndex );
		}
		else if ( ( strncmp(ac_line, "Tr ", 3 ) == 0 ) )
		{ 
		    //Warning: We simply use transparency for refractive color
            //if refractive index != 1.0 && no transparency => set refractive color to 1.0
		    sscanf( ac_line, "Tr %f", &rr );
		    if ( actShader != NULL )
		        actShader->setRefractiveColor( ColorType( rr ) );
		}
		//Alternative version of alpha value
		else if ( ( strncmp(ac_line, "d ", 2 ) == 0 ) )
		{ 
		    //Warning: We simply use transparency for refractive color
		    sscanf( ac_line, "d %f", &rr );
		    if ( actShader != NULL )
		        actShader->setRefractiveColor( ColorType( rr ) );
		}
    }
    
	//Not to forget to add the last Shader to the MaterialList
	if( actShader != NULL )
		materialList->push_back( actShader );
}


TriangleSoup::TriangleSoup( const char* sFilename, std::vector<Shader *>   *materialList ) : 
                     m_bAABBComputed(false),
	                 m_selected(false),
	                 m_drawnThisCycle(false)
{

    m_clVertexList.clear();
	m_clNormalList.clear();
	m_clTriangleList.clear();
	m_clAABBox.Reset();
	
	char	ac_line[256];

	std::ifstream istr_in( sFilename );
	if ( !istr_in.is_open() )
	{
		std::cerr << "could not read " << sFilename << std::endl;
	}

    unsigned int vertexcounter = 0;
    unsigned int facecounter = 0;
	std::cerr << " reading " << sFilename << std::endl;	
	while( istr_in && ( !istr_in.eof() ) && istr_in.getline( ac_line, 256 ) )
	{

		// comment
		if ( ac_line[0] == '#' ) continue;

		// vertex
		else if ( strncmp(ac_line, "v ", 2 ) == 0 )
		{
		    vertexcounter++;
		}
		// face
		else if ( strncmp( ac_line, "f ", 2) == 0)
		{
			int	i3VI[4] = {0, 0, 0, 0};
			int	i3TI[4] = {0, 0, 0, 0};
			int	i3NI[4] = {0, 0, 0, 0};
            //Quadrangel: add 2 triangles
			if ( 12 == sscanf( ac_line, "f %d//%d//%d %d//%d//%d %d//%d//%d %d//%d//%d", &i3VI[0], &i3TI[0], &i3NI[0], 
				                                                                       &i3VI[1], &i3TI[1], &i3NI[1], 
																	                   &i3VI[2], &i3TI[2], &i3NI[2],
																	                   &i3VI[3], &i3TI[3], &i3NI[3] ) )
			{			
		        facecounter++;					
			}
			else if ( 8 == sscanf( ac_line, "f %d//%d %d//%d %d//%d %d//%d", &i3VI[0], &i3NI[0], 
				                                                            &i3VI[1], &i3NI[1], 
																	        &i3VI[2], &i3NI[2],
																	        &i3VI[3], &i3NI[3] ) )
			{						
				facecounter++;		
			}
			else if ( 8 == sscanf( ac_line, "f %d/%d %d/%d %d/%d %d/%d", &i3VI[0], &i3NI[0], 
				                                                         &i3VI[1], &i3NI[1], 
																	     &i3VI[2], &i3NI[2],
																	     &i3VI[3], &i3NI[3] ) )
			{								
				facecounter++;
			}
			else if ( 4 == sscanf( ac_line, "f %d %d %d %d", &i3VI[0], &i3VI[1], &i3VI[2], &i3VI[3] ) )
			{
			    facecounter++;
			}
            
		    facecounter++;
		}
	}

    istr_in.close();
	
	m_clVertexList.reserve( vertexcounter );
	m_clNormalList.reserve( vertexcounter );
	m_clTriangleList.reserve( facecounter );
	m_clActMaterialList.clear();

	unsigned int ui_normals_created = 0;
	unsigned int actFace = 0;

    std::list<MaterialNode> materialNodes;
    materialNodes.clear();

    TriangleSoup *actTriangleSoup = NULL;

    std::ifstream istr2_in( sFilename );

    ActMaterial *newActMaterial = NULL;

	std::cerr << " reading " << sFilename << std::endl;	
	while( istr2_in && ( !istr2_in.eof() ) && istr2_in.getline( ac_line, 256 ) )
	{

		// comment
		if ( ac_line[0] == '#' ) continue;
		
		//Load Material File
		if ( strncmp(ac_line, "mtllib ", 7 ) == 0 )
		{
			char filename[1024];
			if ( sscanf( ac_line, "mtllib %s", filename ) )
			{
				loadMaterials( filename, materialList, &materialNodes );
			}
		}	
		
		//New Material => push New Material to material List 
		if ( strncmp(ac_line, "usemtl ", 7 ) == 0 )
		{
		        if ( newActMaterial != NULL )
		        {
		            newActMaterial->m_stopFace = actFace;
		            m_clActMaterialList.push_back( newActMaterial );    
		        }
    		     
    		    newActMaterial = new ActMaterial();
		    	newActMaterial->m_beginFace = actFace + 1;
   
                bool foundmaterial = false;

			    char matname[1024];
			    if ( sscanf( ac_line, "usemtl %s", matname ) )
			    {
			        for( std::list<MaterialNode>::iterator it = materialNodes.begin();
			            it != materialNodes.end(); ++it )
			            {
			                if ( !strncmp( matname, (*it).m_name, strlen( matname ) ) )
			                {
			                    newActMaterial->m_shader = (*it).m_shader;
			                    foundmaterial = true;
                                break;
			                }
			            }
			    }
                if( !foundmaterial )
                {
                    delete newActMaterial;
                    newActMaterial = NULL;
                }
		}

		// vertex
		else if ( strncmp(ac_line, "v ", 2 ) == 0 )
		{
			float fX, fY, fZ;
			if ( sscanf( ac_line, "v %f %f %f", &fX, &fY, &fZ ) )
			{
				m_clVertexList.push_back( VectorType3( fX, fY, fZ ) );
			}
		}
		// vertex normals
		else if ( strncmp( ac_line, "vn ", 3 ) == 0 )
		{
			float fN1, fN2, fN3;
			if ( sscanf( ac_line, "vn %f %f %f", &fN1, &fN2, &fN3 ) )
			{
				m_clNormalList.push_back( VectorType3( fN1, fN2, fN3 ) );
			}
		}
		// face
		else if ( strncmp( ac_line, "f ", 2) == 0)
		{
			int	i3VI[4] = {0, 0, 0, -1};
			int	i3NI[4] = {0, 0, 0, -1};
			int i3TI[4] = {0, 0, 0, -1};

            //Quadrangel
			if ( 12 == sscanf( ac_line, "f %d//%d//%d %d//%d//%d %d//%d//%d %d//%d//%d", &i3VI[0], &i3TI[0], &i3NI[0], 
				                                                                       &i3VI[1], &i3TI[1], &i3NI[1], 
																	                   &i3VI[2], &i3TI[2], &i3NI[2],
																	                   &i3VI[3], &i3TI[3], &i3NI[3] ) )
			{								
			}
            //Quadrangel
			else if ( 12 == sscanf( ac_line, "f %d/%d/%d %d/%d/%d %d/%d/%d %d/%d/%d", &i3VI[0], &i3TI[0], &i3NI[0], 
				                                                                       &i3VI[1], &i3TI[1], &i3NI[1], 
																	                   &i3VI[2], &i3TI[2], &i3NI[2],
																	                   &i3VI[3], &i3TI[3], &i3NI[3] ) )
			{								
			}
			else if ( 9 == sscanf( ac_line, "f %d//%d//%d %d//%d/%d %d//%d//%d", &i3VI[0], &i3TI[0], &i3NI[0], 
				                                                       &i3VI[1], &i3TI[1], &i3NI[1], 
																	   &i3VI[2], &i3TI[2], &i3NI[2] ) )
			{								
			}
			else if ( 9 == sscanf( ac_line, "f %d/%d/%d %d/%d/%d %d/%d/%d", &i3VI[0], &i3TI[0], &i3NI[0], 
				                                                       &i3VI[1], &i3TI[1], &i3NI[1], 
																	   &i3VI[2], &i3TI[2], &i3NI[2] ) )
			{								
			}
            //Quadrangel
			else if ( 8 == sscanf( ac_line, "f %d//%d %d//%d %d//%d %d//%d", &i3VI[0], &i3NI[0], 
				                                                            &i3VI[1], &i3NI[1], 
																	        &i3VI[2], &i3NI[2],
																	        &i3VI[3], &i3NI[3] ) )
			{						
				++ui_normals_created;		
			}
            //Quadrangel
			else if ( 8 == sscanf( ac_line, "f %d/%d %d/%d %d/%d %d/%d", &i3VI[0], &i3NI[0], 
				                                                         &i3VI[1], &i3NI[1], 
																	     &i3VI[2], &i3NI[2],
																	     &i3VI[3], &i3NI[3] ) )
			{								
				++ui_normals_created;
			}
			else if ( 6 == sscanf( ac_line, "f %d//%d %d//%d %d//%d", &i3VI[0], &i3NI[0], 
				                                                   &i3VI[1], &i3NI[1], 
																   &i3VI[2], &i3NI[2] ) )
			{
				++ui_normals_created;
			}
			else if ( 6 == sscanf( ac_line, "f %d/%d %d/%d %d/%d", &i3VI[0], &i3NI[0], 
				                                                   &i3VI[1], &i3NI[1], 
																   &i3VI[2], &i3NI[2] ) )
			{
				++ui_normals_created;
			}
			//Quadrangel
			else if ( 4 == sscanf( ac_line, "f %d %d %d %d", &i3VI[0], &i3VI[1], &i3VI[2], &i3VI[3] ) )
			{
			}
			else if ( 3 == sscanf( ac_line, "f %d %d %d", &i3VI[0], &i3VI[1], &i3VI[2] ) )
			{
			}
			//obj-indices are 1-based
			i3VI[0] -= 1; i3VI[1] -= 1; i3VI[2] -= 1; i3VI[3] -= 1;
			i3NI[0] -= 1; i3NI[1] -= 1; i3NI[2] -= 1; i3NI[3] -= 1;

			m_clTriangleList.push_back( new Triangle( this, i3VI, i3NI, m_clTriangleList.size() ) );
			
			//Quad-Case: Make second Triangle
			if ( i3VI[3] >= 0 )
			{
			    int	i3V[3];
			    int	i3N[3];
			    i3V[0] = i3VI[0]; i3V[1] = i3VI[2]; i3V[2] = i3VI[3];
			    i3N[0] = i3NI[0]; i3N[1] = i3NI[2]; i3N[2] = i3NI[3];
			    m_clTriangleList.push_back( new Triangle( this, i3V, i3N, m_clTriangleList.size() ) );
			}
			actFace++;
		}
	}

	if ( newActMaterial != NULL )
	{
		newActMaterial->m_stopFace = actFace;
		m_clActMaterialList.push_back( newActMaterial );    
	}

	istr2_in.close();

	std::cout << "vertices:        " << int(m_clVertexList.size()) << std::endl;
	std::cout << "normals:         " << int(m_clNormalList.size()) << std::endl;
	std::cout << "normals created: " << ui_normals_created << std::endl;
	std::cout << "faces:           " << int(m_clTriangleList.size()) << std::endl;

	m_bAABBComputed = false;
	computeBoundingBox();
	std::cout << "min:   " << m_clAABBox.getMin() << std::endl;
	std::cout << "max:   " << m_clAABBox.getMax() << std::endl;
}


//-----------------------------------------------------------------------------

void 
TriangleSoup::computeBoundingBox()
{
	if (m_bAABBComputed) return;
	// compute bounding box
	for ( unsigned int i_index= 0; i_index < m_clVertexList.size(); ++i_index )
	{
		m_clAABBox.extendBy( m_clVertexList[i_index] );
	}

	m_bAABBComputed = true;
}

//-----------------------------------------------------------------------------

void 
TriangleSoup::moveTo( const VectorType3& v3Location )
{
	computeBoundingBox();

	VectorType3 v3_center = m_clAABBox.getCenter();
	VectorType3 v3_diff   = v3Location - v3_center;
	
	translate( v3_diff );
}

//-----------------------------------------------------------------------------

void 
TriangleSoup::translate( const VectorType3& v3Translation )
{
	for ( unsigned int i_index = 0; i_index  < m_clVertexList.size(); ++i_index )
	{
		m_clVertexList[i_index] += v3Translation;
	}
	computeBoundingBox();
}


void 
TriangleSoup::translateRelative( const VectorType3& v3Translation )
{
    translate( v3Translation - m_lastPostition );
    m_lastPostition = v3Translation;
}


//-----------------------------------------------------------------------------

void 
TriangleSoup::scale( const VectorType3& v3scale )
{
	for ( unsigned int i_index = 0; i_index  < m_clVertexList.size(); ++i_index )
	{
		m_clVertexList[i_index] *=  v3scale;		
	}
	computeBoundingBox(); 
}

//-----------------------------------------------------------------------------

void
TriangleSoup::unitize()
{
	computeBoundingBox();

	VectorType3 v3_size = m_clAABBox.getSize();
	VectorType3 v3_scale(1.0/v3_size[0], 1.0/v3_size[1], 1.0/v3_size[2]);
    
	scale( v3_scale );
}

//-----------------------------------------------------------------------------

void 
TriangleSoup::unitizeMax()
{
	computeBoundingBox();

	VectorType3 v3_size = m_clAABBox.getSize();
	
#ifdef WIN32
    double    r_scale = max(max(v3_size[0], v3_size[1]), v3_size[2]) ;
#else
    double    r_scale = std::max(std::max(v3_size[0], v3_size[1]), v3_size[2]);
#endif
	
	
	scale( VectorType3::vectorized( 1.0/r_scale ) );
}

//-----------------------------------------------------------------------------

void
TriangleSoup::updateNormals()
{

}



void TriangleSoup::addTrianglesToScene( SurfaceList *scene )
{
    for ( TriangleListType::iterator it = m_clTriangleList.begin(); it != m_clTriangleList.end(); ++it )
    {
        scene->addSurface( (*it) );
    }
}

void TriangleSoup::renderGL()
{
    if( !m_drawnThisCycle )
    {
    
        m_drawnThisCycle = true;
        //Show object in OpenGLMode
        glPushMatrix();
        
        
	    if( m_selected )
        {
            GLfloat mat_ambient[] = { 1, 1, 1 };
            glMaterialfv( GL_FRONT, GL_AMBIENT, mat_ambient );
        }
        else
            m_shader->shadeGL();
         
        ActMaterial *actmat = NULL;
        if ( m_clActMaterialList.size() > 0 )
            actmat = m_clActMaterialList[0];
        int matcounter = 0;
        
        glBegin( GL_TRIANGLES );
        for ( unsigned int i = 0; i  < m_clTriangleList.size(); i++ )
        {
            if( !m_selected && actmat != NULL && i == actmat->m_beginFace )
            {
                actmat->m_shader->shadeGL();
                matcounter++;
                if ( matcounter < m_clActMaterialList.size() ) 
                    actmat = m_clActMaterialList[matcounter];
            }
            	
            bool setnormal = false;
            for ( unsigned int j = 0; j < 3; j++ )
            {
                VectorType3 vert = m_clVertexList[m_clTriangleList[i]->getVertexIndex(j)];
                glVertex3f( vert[0], vert[1], vert[2] );
                if ( m_clTriangleList[i]->getNormalIndex(j) != -1 )
                {
                    VectorType3 norm = m_clNormalList[m_clTriangleList[i]->getNormalIndex(j)];
                    glNormal3f( norm[0], norm[1], norm[2] );
                    setnormal = true;
                }
            }
            if( !setnormal )
            {
                VectorType3 v3_point0 = m_clVertexList[m_clTriangleList[i]->getVertexIndex(0)];
    	        VectorType3 v3_point1 = m_clVertexList[m_clTriangleList[i]->getVertexIndex(1)];
	            VectorType3 v3_point2 = m_clVertexList[m_clTriangleList[i]->getVertexIndex(2)];

	            VectorType3 u = v3_point1-v3_point0;
	            VectorType3 v = v3_point2-v3_point0;
	            VectorType3 n = u % v;	
	            n.normalize();
        	
                glNormal3f( n[0], n[1], n[2] );
            }
        }
        glEnd();
        glPopMatrix();
    }
}


const Shader *TriangleSoup::getShader(int triangleIndex) const
{
    const Shader *actShader = m_shader;
    for( ActMaterialListType::const_iterator it = m_clActMaterialList.begin();
        it != m_clActMaterialList.end(); ++it )
        {
            if ( (*it)->m_beginFace <= triangleIndex )
                actShader = (*it)->m_shader;
            else
                break;
        }
    
    return actShader;
};